#!/usr/bin/python3
from enum import IntEnum
import dataclasses
import argparse
from turtle import Turtle
from io import BytesIO
import colorsys

# PIL needs ghostscript for EPS processing
from PIL import Image
import numpy as np


class FractalException(Exception):
    pass


class CommonAngles(IntEnum):
    Triangle = 120
    Square = 90
    Hexagon = 60
    Octagon = 45


@dataclasses.dataclass
class FractalSpec:
    axiom: str
    rules: dict
    angle_deg: int
    iterations_warn: int


LSYSTEMS_NAMED = {
    "QUADRATIC_KOCH_ISLAND": FractalSpec(
        "F+F+F+F",
        {
            "F": "F-F+F+FFF-F-F+F",
        },
        CommonAngles.Square,
        3,
    ),
    "TILES": FractalSpec(
        "F+F+F+F",
        {
            "F": "FF+F-F+F+FF",
        },
        CommonAngles.Square,
        4,
    ),
    "QUADRATIC_SNOWFLAKE": FractalSpec(
        "F--F",
        {
            "F": "F-F+F+F-F",
        },
        CommonAngles.Square,
        5,
    ),
    "KOCH_SNOWFLAKE": FractalSpec(
        "F--F--F",
        {
            "F": "F+F--F+F",
        },
        CommonAngles.Hexagon,
        5,
    ),
    "CROSS": FractalSpec(
        "F--F--F",
        {
            "F": "F+FF++F+F",
        },
        CommonAngles.Square,
        5,
    ),
    "CRYSTAL": FractalSpec(
        "F+F+F+F",
        {
            "F": "FF+F++F+F",
        },
        CommonAngles.Square,
        5,
    ),
    "LEVY_C_CURVE": FractalSpec(
        "F",
        {
            "F": "+F--F+",
        },
        CommonAngles.Octagon,
        10,
    ),
    "SIERPINSKI_ARROWHEAD": FractalSpec(
        "YF",
        {
            "X": "YF+XF+Y",
            "Y": "XF-YF-X",
        },
        CommonAngles.Hexagon,
        7,
    ),
    "SIERPINSKI_SIEVE": FractalSpec(
        "FXF--FF--FF",
        {
            "F": "FF",
            "X": "--FXF++FXF++FXF--",
        },
        CommonAngles.Hexagon,
        7,
    ),
    "RINGS": FractalSpec(
        "F+F+F+F",
        {
            "F": "FF+F+F+F+F+F-F",
        },
        CommonAngles.Square,
        4,
    ),
    "PENTAPLEXITY": FractalSpec(
        "F++F++F++F++F",
        {
            "F": "F++F++F+++++F-F++F",
        },
        36,
        5,
    ),
    "HILBERT_CURVE": FractalSpec(
        "L",
        {
            "L": "+RF-LFL-FR+",
            "R": "-LF+RFR+FL-",
        },
        CommonAngles.Square,
        5,
    ),
    "MOORE_CURVE": FractalSpec(
        "LFL-F-LFL",
        {
            "L": "+RF-LFL-FR+",
            "R": "-LF+RFR+FL-",
        },
        CommonAngles.Square,
        7,
    ),
    "TRIANGLE": FractalSpec(
        "F+F+F",
        {
            "F": "F-F+F",
        },
        CommonAngles.Triangle,
        8,
    ),
    "HILBERT_CURVE2": FractalSpec(
        "X",
        {
            "X": "XFYFX+F+YFXFY-F-XFYFX",
            "Y": "YFXFY-F-XFYFX+F+YFXFY",
        },
        CommonAngles.Square,
        5,
    ),
    "32_SEGMENT_CURVE": FractalSpec(
        "F+F+F+F",
        {
            "F": "-F+F-F-F+F+FF-F+F+FF+F-F-FF+FF-FF+F+F-FF-F-F+FF-F-F+F+F-F+",
        },
        CommonAngles.Square,
        3,
    ),
    "THREE_DRAGON_CURVE": FractalSpec(
        "FX+FX+FX",
        {
            "X": "X+YF+",
            "Y": "-FX-Y",
        },
        CommonAngles.Square,
        6,
    ),
}


def colourGenerator(hue_base: float = 0.0,
                    sat_base: float = 0.5,
                    val_base: float = 1.0,
                    hue_step: float = 0.04):
    """
    Generator which steps through color via hue cycling
    """
    sat_base = np.clip(0, 1, sat_base)
    hue_base = np.clip(0, 1, hue_base)
    val_base = np.clip(0, 1, val_base)
    while True:
        color = colorsys.hsv_to_rgb(hue_base, sat_base, val_base)
        yield color
        hue_base += hue_step
        hue_base = hue_base % 1


# code heavily stolen from here
# https://elc.github.io/posts/plotting-fractals-step-by-step-with-python/
# this also looks interesting
# https://colab.research.google.com/github/paulgb/bitaesthetics/blob/master/notebooks/Fractal%20Generation%20with%20L-Systems.ipynb#scrollTo=JRMtBLDFVx_Y


class LSystem(object):

    CANV_WIDTH = 1024
    CANV_HEIGHT = 1024

    def __init__(self, axiom: str, rules: dict, angle: int):
        self.axiom = axiom
        self.rules = rules
        self.angle = angle

        self._frames = []

        # Initialize turtle
        self._turtle = Turtle()
        self._turtle.degrees()  # Require degrees
        self._turtle.speed(0)
        self._turtle.pensize(2)
        self._turtle.hideturtle()

        sc = self._turtle.getscreen()
        sc.setup(self.CANV_WIDTH, self.CANV_HEIGHT)  # Set render canvas
        sc.colormode(cmode=1.0)  # Set RGB color mode to use float components
        sc.bgcolor((.0, .0, .0))  # Set background to black
        self._sc = sc  # Preserve screen for exporting renders

    @staticmethod
    def createFractal(axiom: str, rules: str, iters: int) -> str:
        """
        Create string of instructions for given iterations by recursing through
        rules starting with initial axiom.
        """
        start_string = axiom
        if iters == 0:
            return axiom

        end_string = None
        for _ in range(iters):
            end_string = "".join(rules[i] if i in rules else i for i in start_string)
            start_string = end_string

        return end_string

    def drawFractal(self, iteration, y_offset=0, x_offset=0, offset_angle=0):
        print(f"Starting iteration {iteration}")
        instructions = self.createFractal(self.axiom, self.rules, iteration)

        # make bigger ones smaller so they fit better
        # Not perfect but acceptable for few iterations
        # XXX: Probably would warrant per-system scaling functions
        scale_distance = min(self.CANV_HEIGHT, self.CANV_WIDTH) / 512 \
            * 160.0 / (np.power((iteration + 1), 1.3) * len(self.axiom) * len(self.rules))

        self._turtle.pendown()

        # draw fractal based on l system
        colours_gen = colourGenerator(0.0, 0.7, 1.0, 0.02)
        for cmd in instructions:
            self._turtle.color(next(colours_gen))
            if cmd == 'F':
                self._turtle.forward(scale_distance)
            elif cmd == '+':
                self._turtle.right(self.angle)
            elif cmd == '-':
                self._turtle.left(self.angle)

        # Save result
        self._preserveCanvasFrame()
        self._turtle.penup()
        self._turtle.home()

        self._turtle.clear()

    def _preserveCanvasFrame(self):
        tkcv = self._sc.getcanvas()
        tkcv.configure(bg="black")
        ps = tkcv.postscript(colormode="color")
        image = Image.open(BytesIO(bytes(ps, "ascii")))
        image.load(transparency=True)
        image.convert("RGBA")
        new_image = Image.new("RGBA", image.size, (0, 0, 0))
        new_image.alpha_composite(image)
        self._frames.append(new_image)

    def saveAnimation(self, filename):
        print("Saving animation")
        assert self._frames  # Should not be empty
        img = self._frames.pop(0)
        img.save(fp=filename, format='WEBP', append_images=self._frames,
                 save_all=True, duration=300, loop=0, allow_mixed=False,
                 lossless=True)
# class LSystem


def main() -> int:
    # Parse arguments
    parser = argparse.ArgumentParser(
        prog="fractal.py",
        description="Fractal drawer.",
        add_help=True,
    )
    parser.add_argument("--fractal", type=str,
                        nargs=1, required=False, action="store",
                        default=["KOCH_SNOWFLAKE"],
                        choices=list(LSYSTEMS_NAMED))
    parser.add_argument("-i", "--iterations", type=int,
                        nargs=1, required=False, action="store")
    parser.add_argument("output", type=str,
                        nargs="?", action="store", default="out")
    args = parser.parse_args()

    iterations = args.iterations
    output_filename = args.output
    print(args.fractal)
    fractal = LSYSTEMS_NAMED[args.fractal.pop()]

    # Process arguments
    if iterations is None:
        iterations = fractal.iterations_warn
    elif iterations > fractal.iterations_warn:
        print(f"Iterations greater than {fractal.iterations_warn} will"
              f"take bloody ages, setting to {fractal.iterations_warn}")
        iterations = fractal.iterations_warn

    # Generate fractal
    lsystem = LSystem(fractal.axiom, fractal.rules, fractal.angle_deg)
    for i in range(iterations):
        lsystem.drawFractal(i, x_offset=(i ** 2) * 10, y_offset=(i ** 2) * 10)
    lsystem.saveAnimation(output_filename)

    return 0


if __name__ == "__main__":
    main()
